
#include <dos.h>
#include <dir.h>
#include <ctype.h>
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define VERSION "1.10"
#define CLEAN	0
#define INFECTED	1
#define MUTATION	2

typedef struct dev_hdr {
	struct dev_hdr far *dh_next;	/* Next device pointer */
	short		dh_attr;	/* Attributes */
	unsigned short	dh_strat;	/* Driver strategy routine */
	unsigned short	dh_inter;	/* Driver interrupt routine */
	char		dh_name [8];	/* Device name */
} DEV_HDR;

typedef struct {
	unsigned char name [8];
	unsigned char ext  [3];
	unsigned char attr;
	unsigned char reserved [8];
	unsigned key;
	unsigned date;
	unsigned time;
	unsigned cluster;
	unsigned long size;
} DIRENTRY;

typedef struct DPB3 {
	char Drive;
	char SubUnit;
	unsigned SectSiz;	/* Bytes per sector */
	char HiSec;		/* Sectors per cluster - 1 */
	char CSS;		/* Log base two of the cluster size */
	unsigned ResSecs;
	char FatCnt;		/* FAT numbers */
	unsigned RootSiz;	/* Number of root directory entries */
	unsigned DataSec;	/* Sector of first data cluster */
	unsigned HiClust;	/* Total clusters */
	char FatSize;		/* Sectors per FAT copy */
	unsigned RootSec;	/* Sector of root directory */
	DEV_HDR far *Driver;
	char Media;
	char Acc;
	struct DPB3 far *Next;
} DPB3;

typedef struct DPB4 {
	char Drive;
	char SubUnit;
	unsigned SectSiz;	/* Bytes per sector */
	char HiSec;		/* Sectors per cluster - 1 */
	char CSS;		/* Log base two of the cluster size */
	unsigned ResSecs;
	char FatCnt;		/* FAT numbers */
	unsigned RootSiz;	/* Number of root directory entries */
	unsigned DataSec;	/* Sector of first data cluster */
	unsigned HiClust;	/* Total clusters */
	unsigned FatSize;	/* Sectors per FAT copy */
	unsigned RootSec;	/* Sector of root directory */
	DEV_HDR far *Driver;
	char Media;
	char Acc;
	struct DPB4 far *Next;
} DPB4;

unsigned fat_type,	/* 1 - 12 bit; 0 - 16 bit */
	 drive, 	/* Drive letter */
	 max_cluster,
	 fat_index,
	 fat_offset,
	 fat_sector,
	 last_sector,
	 files_tested,
	 files_infected,
	 files_cured,
	 cure,
	 found;

unsigned long body_sector;

unsigned char fat_buffer [1024], virus_body [1024];

unsigned keys [16];

DPB4 disk_info;

void clean_dir (char *lpath, unsigned cluster);

#define NUM_OFFS	15	/* signature length */
static struct mem_sign {
	unsigned	strat_offs;
	unsigned	vir_len;
	unsigned	num_offs;
	unsigned	offs [NUM_OFFS];
	unsigned	sum;
	int		cnd_offs1, cnd_offs2;
	unsigned char	cnd_new2;
} mem_sign [] = {
	{	/* original virus */
		0x14B,
		0x3EB,	/* 1024-21 */
		15,
		{
			0x334, 0x335, 0x33F, 0x340, 0x365, 0x394, 0x395, 0x3B2,
			0x3B3, 0x3B4, 0x3B5, 0x3B7, 0x3B8, 0x3B9, 0x3BA
		},
		0x95A1,
		0x3B1, 0x3C0,
		       0xCB	/* retf */
	}
};

void error (int err_num)
{
	fprintf (stderr, "DIR2CLR: ");
	switch (err_num) {
		case 1:
			fprintf (stderr, "Can't read sector on drive %c:.\n", drive);
			break;
		case 2:
			fprintf (stderr, "Can't write sector on drive %c:.\n", drive);
			break;
		case 3:
			fprintf (stderr, "Can't read FAT on drive %c:.\n", drive);
			break;
		case 4:
			fprintf (stderr, "Can't write FAT on drive %c:.\n", drive);
			break;
		case 5:
			fprintf (stderr, "Bad drive letter %c.\n", drive);
			return;
		case 6:
			fprintf (stderr, "Invalid drive %c:.\n", drive);
			return;
		case 7:
			fprintf (stderr, "Incompatible sector size on drive %c:.\n", drive);
			return;
		case 8:
			fprintf (stderr, "usage: DIR2CLR [-n] drive...\n\t-n - Don't cure the infected files\n");
			break;
		case 9:
			fprintf (stderr, "Mutation found active in memory!\n");
	}
	exit (2);
}

void mk_dinfo (DPB3 far *disk_info3)
{
	if (_osmajor > 3)
		movedata (FP_SEG (disk_info3), FP_OFF (disk_info3),
			  FP_SEG (&disk_info), FP_OFF (&disk_info),
			  sizeof (DPB4));
	else {
		disk_info.Drive   = disk_info3->Drive;
		disk_info.SubUnit = disk_info3->SubUnit;
		disk_info.SectSiz = disk_info3->SectSiz;
		disk_info.HiSec   = disk_info3->HiSec;
		disk_info.CSS	  = disk_info3->CSS;
		disk_info.ResSecs = disk_info3->ResSecs;
		disk_info.FatCnt  = disk_info3->FatCnt;
		disk_info.RootSiz = disk_info3->RootSiz;
		disk_info.DataSec = disk_info3->DataSec;
		disk_info.HiClust = disk_info3->HiClust;
		disk_info.FatSize = disk_info3->FatSize;
		disk_info.RootSec = disk_info3->RootSec;
		disk_info.Driver  = disk_info3->Driver;
		disk_info.Media   = disk_info3->Media;
		disk_info.Acc	  = disk_info3->Acc;
		disk_info.Next	  = (DPB4 *) disk_info3->Next;
	}
}

void disk_io (int do_write, int num_secs, unsigned long first_sec, void *buffer, int err_code)
{
	if (do_write) {
		if (abswrite (drive - 'A', num_secs, first_sec, buffer))
			error (err_code);
	} else {
		if (absread (drive - 'A', num_secs, first_sec, buffer))
			error (err_code);
	}
}

void read_sector (unsigned long sector, DIRENTRY *dir_buf)
{
	disk_io (0, 1, sector, dir_buf, 1);
}

void write_sector (unsigned long sector, DIRENTRY *dir_buf)
{
	disk_io (1, 1, sector, dir_buf, 2);
}

char *fname (DIRENTRY *entry)
{
	register i, j;
	static char fnames [15];

	for (j = i = 0; i < 8; i++, j++)
		if (entry->name [i] == ' ')
			break;
		else
			fnames [j] = entry->name [i];
	if (! (entry->attr & 0x10))
		fnames [j++] = '.';
	for (i = 0; i < 3; i++, j++)
		if (entry->ext [i] == ' ')
			break;
		else
			fnames [j] = entry->ext [i];
	fnames [j] = '\0';
	return (fnames);
}

void make_path (char *path, char *oldpath, DIRENTRY *entry)
{
	register i, j;

	strcpy (path, oldpath);
	for (i = 0, j = strlen (path); i < 8; i++, j++)
		if (entry->name [i] == ' ')
			break;
		else
			path [j] = entry->name [i];
	for (i = 0; i < 3; i++, j++)
		if (entry->ext [i] == ' ')
			break;
		else
			path [j] = entry->ext [i];
	path [j] = '\0';
	strcat (path, "\\");
}

unsigned fat_entry (unsigned cluster)
{
	int odd;

	odd = cluster & 1;
	fat_index = (cluster >> fat_type) + cluster;
	fat_sector = fat_index >> 9;
	fat_offset = fat_index % 512;
	if (fat_sector != last_sector) {
		disk_io (0, 1 + fat_type, fat_sector + disk_info.ResSecs, fat_buffer, 3);
		last_sector = fat_sector;
	}
	cluster = * (unsigned *) (fat_buffer + fat_offset);
	if (fat_type) {
		if (odd)
			cluster >>= 4;
		else
			cluster &= 0x0FFF;
		if (fat_type && cluster >= 0x0FF0)
			cluster |= 0xF000;
	}
	return cluster;
}

void free_entry (unsigned cluster)
{
	fat_entry (cluster);
	* (unsigned *) (fat_buffer + fat_index % 512) &= (fat_type) ? ((cluster & 1) ? 0x000F : 0xF000) : 0;
	disk_io (1, 1 + fat_type, fat_sector + disk_info.ResSecs, fat_buffer, 4);
}

unsigned clean_entry (DIRENTRY *entry, char *path, unsigned offs)
{
	unsigned f = 0;

	if ((memcmp ("COM", entry->ext, 3) == 0 || memcmp ("EXE", entry->ext, 3) == 0) &&
	    (entry->attr & 0x1C) == 0) {
		if (entry->name [0] != 0xE5) {
			fprintf (stdout, "\r%c:%s%s", drive, path, fname (entry));
			clreol ();
			files_tested++;
		}
		if (entry->cluster == max_cluster)
			if (! cure) {
				if (entry->name [0] != 0xE5) {
					fprintf (stdout, " is infected!\n");
					files_infected++;
				}
			} else {
				entry->cluster = entry->key ^ keys [offs];
				entry->key = 0;
				f = 1;
				if (entry->name [0] != 0xE5) {
					files_cured++;
					fprintf (stdout, " - cured\n");
					files_infected++;
				}
			}
       }
       return (f);
}

void clean_dir_cluster (unsigned num_secs, unsigned long start_sec,
			char *lpath, DIRENTRY *dir_buf, char *path)
{
	register i, j, f;

	for (j = 0; j < num_secs; j++) {
		read_sector (start_sec + j, &dir_buf [0]);
		for (f = i = 0; i < 16; i++) {
			if (dir_buf [i].name [0] == '\0') {
				if (f)
					write_sector (start_sec + j, &dir_buf [0]);
				return;
			}
			if (dir_buf [i].attr & 0x10 &&
			    dir_buf [i].name [0] != '.' &&
			    dir_buf [i].name [0] != 0xE5) {
				make_path (path, lpath, &dir_buf [i]);
				clean_dir (path, dir_buf [i].cluster);
			} else if (clean_entry (&dir_buf [i], lpath, i))
				f = 1;
		}
		if (f)
			write_sector (start_sec + j, &dir_buf [0]);
	}
}

void clean_dir (char *lpath, unsigned cluster)
{
	unsigned long sector;
	char path [MAXPATH];
	DIRENTRY dir_buf [16];

	do {
		sector = ((long) cluster - 2) * disk_info.HiSec + disk_info.DataSec;
		clean_dir_cluster (disk_info.HiSec, sector, lpath, dir_buf, path);
		cluster = fat_entry (cluster);
	} while (cluster < 0xFFF0);
}

void clean_drive (void)
{
	unsigned dir_size;
	DIRENTRY dir_buf [16];
	char path [MAXPATH];

	found = files_tested = files_infected = files_cured = 0;
	/* Clean the root directory */
	dir_size = ((disk_info.RootSiz << 5) + 511) >> 9;
	clean_dir_cluster (dir_size, (unsigned long) disk_info.RootSec, "\\", dir_buf, path);
	fprintf (stdout, "\r");
	clreol ();
	fprintf (stdout, "\n%u file(s) tested; %u file(s) infected; %u file(s) cured.\n\n",
			  files_tested, files_infected, files_cured);
	if (cure) {
		free_entry (max_cluster++);
		if (disk_info.HiSec < 2)
			free_entry (max_cluster);
		memset(virus_body, '\0', sizeof(virus_body));
		disk_io (1, 2, body_sector, virus_body, 2);
	}
}

unsigned chk_sum (int mutno)
{
	register unsigned i, sum;

	for (i = 0; i < mem_sign [mutno].num_offs; virus_body [mem_sign [mutno].offs [i++]] = 0);
	for (i = mem_sign [mutno].vir_len; i < sizeof (virus_body); virus_body [i++] = 0 );
	for (i = sum = 0; i < sizeof (virus_body) / sizeof (unsigned); sum ^= *((unsigned int *) virus_body + i++));
	return sum;
}

check_inf (void)
{
	register unsigned i, magic;
	int status = CLEAN;
	static unsigned char signature [] = {
		0xBC, 0x00, 0x06, 0xFF
	};

	last_sector = 0;
	max_cluster = disk_info.HiClust + 1;
	do
		i = fat_entry (--max_cluster);
	while (i == 0xFFF7);
	body_sector = disk_info.DataSec + ((unsigned long) max_cluster - 2) * disk_info.HiSec;
	if (disk_info.HiSec < 2) {
		body_sector--;
		max_cluster--;
	}
	disk_io (0, 2, body_sector, virus_body, 1);
	if (0xFFF8 <= i && i <= 0xFFFE || memcmp (signature, virus_body, sizeof (signature)) == 0)
		status = MUTATION;
	magic = * (unsigned *) (virus_body + 0x33F);
	for (i = 0; i < 16; i++) {
		keys [i] = magic;
		magic = (magic << 1) + ((magic & 0x8000) != 0);
	}
	if (chk_sum(0) == mem_sign[0].sum)
		status = INFECTED;
	return status;
}

void mem_clr (void)
{
	register i;
	DEV_HDR huge *vir_dev;
	unsigned char huge *vir_strat, huge *p;

	_AH = 0x52;
	__int__ (0x21);
	mk_dinfo (* (DPB3 far * far *) MK_FP (_ES,_BX));
	vir_dev = disk_info.Driver;
	vir_strat = MK_FP (FP_SEG (vir_dev), vir_dev->dh_strat);
	for (i = 0; i < sizeof (mem_sign) / sizeof (struct mem_sign); i++) {
		movedata (FP_SEG (vir_strat - mem_sign [i].strat_offs),
			  FP_OFF (vir_strat - mem_sign [i].strat_offs),
			  FP_SEG (virus_body), FP_OFF (virus_body),
			  mem_sign [i].vir_len);
		if (chk_sum (i) == mem_sign [i].sum) {
			fprintf (stdout, "Virus found in memory and deactivated.\n\n");
			vir_dev->dh_strat += mem_sign [i].cnd_offs1 - mem_sign [i].strat_offs;
			p = vir_strat + (mem_sign [i].cnd_offs2 - mem_sign [i].strat_offs);
			*p = mem_sign [i].cnd_new2;
		}
	}
}

main (int argc, char **argv)
{
	unsigned offs;

	fprintf (stdout,
"\nDir2Clr: Program to search and cure the DIR II virus.\n\
Copyright (C) 1991 by Vesselin Bontchev. Version %s.\n\
Laboratory of Computer Virology, Bulgarian Academy of Sciences,\n\
Sofia, ul. Acad. G. Bontchev, bl. 8, rm. 104, tel. +359-2-719212\n\n", VERSION);
	if (argc < 2)
		error (8);
	cure = 1;
	if (0 == stricmp (*++argv, "-n")) {
		cure = 0;
		argv++;
	}
	mem_clr ();
	for (; *argv != NULL; argv++) {
		drive = toupper (**argv);
		if (! isupper (drive)) {
			error (5);
			continue;
		}
		__emit__ (0x1E); /* push ds */
		_DL = drive - 'A' + 1;
		_AH = 0x32;
		__int__ (0x21);
		__emit__ (0x1E, 0x07, 0x1F); /* push ds; pop es; pop ds */
		if (_AL) {
			error (6);
			continue;
		}
		offs = _BX;
		pokeb (_ES, offs + 0x17 + (_osmajor >= 4), 0xFF);
		_AH = 0x0D;
		__int__ (0x21);
		__emit__ (0x1E); /* push ds */
		_DL = drive - 'A' + 1;
		_AH = 0x32;
		__int__ (0x21);
		__emit__ (0x1E, 0x07, 0x1F); /* push ds; pop es; pop ds */
		if (_AL) {
			error (6);
			continue;
		}
		mk_dinfo (MK_FP (_ES, _BX));
		disk_info.HiSec++;
		if (disk_info.SectSiz != 512) {
			error (7);
			continue;
		}
		fat_type = disk_info.HiClust <= 0xFF0;
		switch (check_inf ()) {
			case INFECTED:
				clean_drive ();
				if (found < 1)
					found = 1;
				break;
			case MUTATION:
				fprintf (stdout, "Drive %c: is infected by a mutation!\n", drive);
				found = 2;
				break;
			default:
				fprintf (stdout, "Drive %c: is not infected.\n", drive);
		}
	}
	return found;
}

